//=============================================================================
// FMMutator.
//
// Author: Francesco Biscazzo
// Date: 2019
// ©copyright Francesco Biscazzo. All rights reserved.
//
// Description: September!
//=============================================================================
class FMMutator extends Mutator
	config(FunkyMoves);

const version = "1r4";
	
var config bool bLogs;

var config bool bCmdInform; // If true the players will be told the command to print all the available commands.
var config String cmdFormatting; // Symbol/s to be put at the start of the message to be interpreted as a command.
var config bool bCmdFormattingMutate; // If true cmdFormatting must be put when using "mutate" command.

var config int maxCPs; // Max amount of CPs that can be spawned. When reached, CPs will start to recycle.
var config bool bTeleportEffect; // If true the GameType's teleport effect will be spawned at target location when teleporting.

event PreBeginPlay() {
	super.PreBeginPlay();
	
	Level.Game.RegisterMessageMutator(self);
	
	if (bLogs)
		Warn("bLogs is set to True");
}

/*
 *	Return the pawn of class pawnClass that contains nameSnippet to the most left in its name.
 *
 *	E.g. With a name snippet of "b" and two pawns called "ab" and "ba", the player called "ba" will be returned.
 */
function Pawn getPawnByName(String nameSnippet, optional class<Pawn> pawnClass) {
	local int prevIndex, currIndex;
    local Pawn pawn, foundPawn;
	
	if (pawnClass == None)
		pawnClass = class'Pawn';
	
    foreach AllActors(class'Pawn', pawn) {
		if ((pawn.class == pawnClass) || ClassIsChildOf(pawn.class, pawnClass)) {
			if (pawn.playerReplicationInfo != None) {
				currIndex = InStr(Caps(pawn.playerReplicationInfo.playerName), Caps(nameSnippet));
				if (currIndex != -1) {
					if (foundPawn == None) {
						foundPawn = pawn;
					} else {
						if (currIndex < prevIndex)
							foundPawn = pawn;
					}
					
					prevIndex = currIndex;
				}
			}
		}
	}
    
	return foundPawn;
}

/*
 *	Return true if the checkpoint was placed. Return false otherwise.
 *	
 *	If there are already maxCPs checkpoints for the actorOwner then recycle them turning the oldest one into the newset.
 */
function bool placeCP(Actor actorOwner, optional bool bInformPlayer) {
	local FMCheckPoint CP, oldestCP;
	local int counter;
	local Inventory inv;
	
	foreach AllActors(class'FMCheckPoint', CP) {
		if (CP.Owner == actorOwner) {
			if (oldestCP == None)
				oldestCP = CP;
			else
				if (CP.timeSeconds < oldestCP.timeSeconds)
					oldestCP = CP;
					
			if (++counter >= maxCPs)
				break;
		}
	}
	
	if (counter < maxCPs)
		CP = Level.Spawn(class'FMCheckPoint', actorOwner);
	else
		CP = oldestCP;
	if (CP != None) {
		CP.timeSeconds = Level.timeSeconds;
		CP.setLocation(actorOwner.Location);
		CP.setRotation(actorOwner.Rotation);
		if (actorOwner.IsA('Pawn'))
			CP.viewRotation = Pawn(actorOwner).viewRotation;
		setCPsDisplay(actorOwner);
		
		if (bInformPlayer && actorOwner.IsA('PlayerPawn'))
			PlayerPawn(actorOwner).ClientMessage("Checkpoint placed!");
			
		return true;
	}
	
	if (bInformPlayer && actorOwner.IsA('PlayerPawn'))
		PlayerPawn(actorOwner).ClientMessage("Checkpoint placed!");
	
	return false;
}

/*
 *	Delete all the checkpoints placed by actorOwner.
 */
function deleteCPs(Actor actorOwner, optional bool bInformPlayer) {
	local FMCheckPoint CP;
	
	foreach AllActors(class'FMCheckPoint', CP)
		if (CP.Owner == actorOwner)
			CP.Destroy();
			
	if (bInformPlayer && actorOwner.IsA('PlayerPawn'))
		PlayerPawn(actorOwner).ClientMessage("Checkpoint deleted!");
}

/*
 *	Return true if actor was successfully moved to target. Return false otherwise.
 *	
 *	@bNext If true actor will be moved next to the target. If false actor will be moved exactly at target location.
 */
function bool teleport(Actor actor, Actor target, optional bool bNoTelefrag, optional bool bNext, optional bool bInformPlayer) {
	local vector loc;
	local rotator viewRotation;
	local bool bFoundStart;
	
	if (bNoTelefrag)
		actor.SetCollision(false, false, false);
	
	loc = target.Location;
	if (bNext)
		loc += 72 * Vector(target.Rotation) + vect(0,0,1) * 15;
		
	bFoundStart = actor.setLocation(loc);
	if (bFoundStart) {
		if (bTeleportEffect)
			Level.Game.PlayTeleportEffect(actor, true, true);
	
		actor.setRotation(target.Rotation);
		if (target.IsA('Pawn'))
			viewRotation = Pawn(target).viewRotation;
		else if (target.IsA('FMCheckPoint'))
			viewRotation = FMCheckPoint(target).viewRotation;
		if (actor.IsA('Pawn'))
			Pawn(actor).ClientSetLocation(loc, viewRotation);
		else if (actor.IsA('FMCheckPoint'))
			FMCheckPoint(actor).viewRotation = viewRotation;
		
		if (bInformPlayer && actor.IsA('PlayerPawn'))
			PlayerPawn(actor).ClientMessage("You have been teleported to the target.");
		
		actor.SetCollision(actor.default.bCollideActors, actor.default.bBlockActors, actor.default.bBlockPlayers);
		
		return true;
	}
	
	if (bInformPlayer && actor.IsA('PlayerPawn'))
		PlayerPawn(actor).ClientMessage("Teleportation failed: Collision conflict on the target location.");
	
	actor.SetCollision(actor.default.bCollideActors, actor.default.bBlockActors, actor.default.bBlockPlayers);
	
	return false;
}

/*
 *	Return the checkpoint with the most recent timestamp placed by actorOwner.
 */
function FMCheckPoint getNewestCP(Actor actorOwner) {
	local FMCheckPoint CP, newestCP;
	
	foreach AllActors(class'FMCheckPoint', CP) {
		if (CP.Owner == actorOwner) {
			if (newestCP == None)
				newestCP = CP;
			else
				if (CP.timeSeconds > newestCP.timeSeconds)
					newestCP = CP;
		}
	}
	
	return newestCP;
}

/*
 *	Hide all the checkpoints while unhiding the one with the most recent timestamp placed by actorOwner.
 */
function setCPsDisplay(Actor actorOwner) {
	local FMCheckPoint CP, newestCP;
	
	foreach AllActors(class'FMCheckPoint', CP) {
		if (CP.Owner == actorOwner) {
			if (newestCP == None)
				newestCP = CP;
			else
				if (CP.timeSeconds > newestCP.timeSeconds)
					newestCP = CP;
					
			CP.bHidden = true;
		}
	}
	
	newestCP.bHidden = false;
}

/*
 *	Return true if actorOwner was successfully moved to the checkpoint with the most recent timestamp placed by actorOwner. Return false otherwise.
 */
function bool teleportToNewestCP(Actor actorOwner, optional bool bInformPlayer) {
	local FMCheckPoint CP;
	
	CP = getNewestCP(actorOwner);
	if (CP != None)
		return teleport(actorOwner, CP, true, false, bInformPlayer);
	else if (actorOwner.IsA('PlayerPawn') && bInformPlayer)
		PlayerPawn(actorOwner).ClientMessage("No checkpoint set!");
		
	return false;
}

/*
 *	Return true if the checkpoint with the most recent timestamp placed by actorOwner was successfully deleted. Return false if actorOwner owns only one checkpoint.
 */
function bool deleteNewestCP(Actor actorOwner, optional bool bInformPlayer) {
	local FMCheckPoint CP, newestCP;
	local int counter;
	
	foreach AllActors(class'FMCheckPoint', CP) {
		if (CP.Owner == actorOwner) {
			if (newestCP == None)
				newestCP = CP;
			else
				if (CP.timeSeconds > newestCP.timeSeconds)
					newestCP = CP;
					
			counter++;
		}
	}
	if ((counter > 1) && (newestCP != None)) {
		newestCP.Destroy();
		getNewestCP(actorOwner).bHidden = false;
		if (actorOwner.IsA('PlayerPawn') && bInformPlayer)
			PlayerPawn(actorOwner).ClientMessage("Your CP has been moved to a previous location.");
			
		return true;
	} else if (counter == 1)
		if (actorOwner.IsA('PlayerPawn') && bInformPlayer)
			PlayerPawn(actorOwner).ClientMessage("Failed to move your CP to a previous location, you are left with 1 checkpoint only!");
	
	return false;
}

function ModifyPlayer(Pawn Other) {
	local FMLoginNotification notification;
	local bool bNotified;

	super.ModifyPlayer(Other);
	
	teleportToNewestCP(Other, false);
	
	if (bCmdInform)
		if (Other.IsA('PlayerPawn')) {
			foreach AllActors(class'FMLoginNotification', notification)
				bNotified = (notification.Owner == Other);
			if (!bNotified) {
				Spawn(class'FMLoginNotification', Other);
				PlayerPawn(Other).ClientMessage("FunkyMoves_v"$version$" - Type "$cmdFormatting$"cmd to display all the available commands.");
			}
		}
}

/*
 *	Return true if pawn was successfully moved to a PlayerStart. Return false otherwise.
 */
function bool teleportToPlayerStart(Pawn pawn, optional bool bInformPlayer) {
	local NavigationPoint startSpot;
	
	startSpot = Level.Game.FindPlayerStart(pawn, 255);
	if (startSpot != None)
		if (teleport(pawn, startSpot, true, false, bInformPlayer))
			return true;
	
	if (pawn.IsA('PlayerPawn') && bInformPlayer)
		PlayerPawn(pawn).ClientMessage("Teleportation failed!");
	
	return false;
}

function Mutate(String MutateString, PlayerPawn Sender) {
	super.Mutate(MutateString, Sender);

	if (!bCmdFormattingMutate)
		MutatorTeamMessage(Sender, Sender, Sender.PlayerReplicationInfo, MutateString, 'MUTATE');
	else if (bCmdFormattingMutate && (Left(MutateString, Len(cmdFormatting)) == cmdFormatting))
		MutatorTeamMessage(Sender, Sender, Sender.PlayerReplicationInfo, Right(MutateString, Len(MutateString) - Len(cmdFormatting)), 'MUTATE');
}

function bool MutatorTeamMessage(Actor Sender, Pawn Receiver, PlayerReplicationInfo PRI, coerce string S, name Type, optional bool bBeep) {
	local bool bCensored;
	
	local Pawn pawnTarget;
	local PlayerPawn senderPP;
    local string playerName, command;
    local int separatorIndex;
	local FMCheckPoint CP;
	
	if (Sender == Receiver) {
		senderPP = PlayerPawn(Sender);
	
		command = S;
		if (((Type != 'MUTATE') && (Left(S, Len(cmdFormatting)) == cmdFormatting)) || (Type == 'MUTATE')) {
			if (Type != 'MUTATE')
				command = Right(S, Len(S) - Len(cmdFormatting));
			if (InStr(command, " ") > 0) {
				separatorIndex = InStr(command, " ");
				command = Mid(command, 0, separatorIndex);
			}
			switch (Caps(command)) {
				case "CMD":
					if (Sender.IsA('PlayerPawn')) {
						senderPP.ClientMessage(" ");
						senderPP.ClientMessage("[FunkyMoves_v"$version$" Commands BEGIN]");
						if (bCmdFormattingMutate)
							senderPP.ClientMessage("NOTE: These commands need the trailing formatting '"$cmdFormatting$"' to be executed through 'mutate' command");
						else
							senderPP.ClientMessage("NOTE: These commands do not need the trailing formatting '"$cmdFormatting$"' to be executed through 'mutate' command");
						senderPP.ClientMessage(cmdFormatting$"cmd - Displays this thing.");
						senderPP.ClientMessage(cmdFormatting$"spawn - Teleports you to a PlayerStart.");
						senderPP.ClientMessage(cmdFormatting$"cp - Creates a checkpoint.");
						senderPP.ClientMessage(cmdFormatting$"prevcp - Moves your CP to a previous location.");
						senderPP.ClientMessage(cmdFormatting$"nocp - Deletes your checkpoint.");
						senderPP.ClientMessage(cmdFormatting$"gocp - Teleports you to your checkpoint.");
						senderPP.ClientMessage(cmdFormatting$"go <player_name> - Teleports you to the first player found with a similar player_name.");
						senderPP.ClientMessage(cmdFormatting$"goview - Teleports you to the viewing player.");
						senderPP.ClientMessage("[FunkyMoves_v"$version$" Commands END]");
						senderPP.ClientMessage(" ");
					}
					
					break;
				case "SPAWN":
					if (Sender.IsA('Pawn'))
						teleportToPlayerStart(Pawn(Sender), true);
					
					break;
				case "CP":
					if (Sender.Physics != PHYS_FALLING)
						placeCP(Sender, true);
					else if (Sender.IsA('PlayerPawn'))
						senderPP.ClientMessage("You cannot place the CP in mid air.");
					
					break;
				case "PREVCP":
					deleteNewestCP(Sender, true);
					
					break;
				case "NOCP":
					deleteCPs(Sender, true);
							
					break;
				case "GOCP":
					teleportToNewestCP(Sender, true);
						
					break;
				case "GO":
					playerName = Mid(S, InStr(S, " ") + 1, Len(S));
					pawnTarget = getPawnByName(playerName, class'PlayerPawn');
					if (pawnTarget != Sender)
						teleport(Sender, pawnTarget, true, false, true);
					else if (Sender.IsA('PlayerPawn'))
						senderPP.ClientMessage("Teleportation failed: No player found!");
					
					break;
				case "GOVIEW":
					if (Sender.IsA('PlayerPawn'))
						if ((senderPP.ViewTarget != None) && senderPP.ViewTarget.IsA('PlayerPawn'))
							teleport(Sender, senderPP.ViewTarget, true, false, true);
					
					break;
			}
		}
	}
	
	return !bCensored && super.MutatorTeamMessage(Sender, Receiver, PRI, S, Type, bBeep);
}

defaultproperties {
	bLogs=False
	bCmdInform=True
	cmdFormatting="/"
	bCmdFormattingMutate=True
	maxCPs=3
	bTeleportEffect=False
	RemoteRole=ROLE_SimulatedProxy
	bAlwaysRelevant=True
}